/***************************************************************************
 *   Copyright (C) 2010 by                                                 *
 *                                                                         *
 *   catWorkX GmbH, Schellerdamm 16, 21079 Hamburg                         *                                                                         
 *   Sebastian Gerdes <sebastian.gerdes_AT_catworkx.de>                    *
 *                                                                         *
 *   This code is part of the Zarafa 2 LDAP tool, which is able to export  *
 *   private and/or public contacts from Zarafa to a LDAP server           * 
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Affero General Public License as        *
 *   published by the Free Software Foundation; either version 3           *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          *
 *   GNU AFFERO GENERAL PUBLIC LICENSE for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU Affero General Public      *
 *   License along with this program.  If not, see                         *
 *   <http://www.gnu.org/licenses/>                                        *
 *                                                                         *
 ***************************************************************************/

package de.catworkx.zarafa.zarafa2ldap;

import java.io.File;
import java.io.FileInputStream;
import java.util.Hashtable;
import java.util.Properties;

import javax.naming.Context;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.BasicAttribute;
import javax.naming.directory.BasicAttributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;

public class Zarafa2Ldap {


	public static void main(String[] args) throws NamingException {
		
		Properties config = new Properties();
		
		try {
			String path = new File(".").getCanonicalPath();
			String configFile = "zarafa2ldap.conf";
			config.load(new FileInputStream(path.concat("/").concat(configFile)));
		} catch (Exception e) {
			System.out.println("Config file zarafa2ldap.conf not found!");
		} 
		
		String ldapserver = config.getProperty("ldapserver");
		String ldapuser = config.getProperty("ldapuser");
		String ldappass = config.getProperty("ldappass");
		String baseDN = config.getProperty("baseDN");
		String sync_privatestore = config.getProperty("sync_privatestore");
		String sync_publicstore = config.getProperty("sync_publicstore");
		
		MapiConnect mapi = new MapiConnect(config.getProperty("soapUrl"));
		
		// LDAP environment using connection pooling
		Hashtable env = new Hashtable();
		env.put(Context.INITIAL_CONTEXT_FACTORY,"com.sun.jndi.ldap.LdapCtxFactory");
		env.put(Context.PROVIDER_URL, ldapserver.concat(baseDN));
		env.put("com.sun.jndi.ldap.connect.pool", "true");
		env.put(Context.SECURITY_AUTHENTICATION, "simple");

		env.put(Context.SECURITY_PRINCIPAL, ldapuser);
		env.put(Context.SECURITY_CREDENTIALS, ldappass);
		
		
		ldapInitialize(env, ldapserver, baseDN);
		ldapCleanup(env, ldapserver, baseDN);
		
		if (sync_privatestore.equalsIgnoreCase("true")) {
			writePrivateContacts(env, ldapserver, baseDN, mapi);
		}
		
		if (sync_publicstore.equalsIgnoreCase("true")) {
			writePublicContacts(env, ldapserver, baseDN, mapi);
		}			
	}

	private static void ldapInitialize (Hashtable env, String ldapserver, String baseDN) throws NamingException {
		
		DirContext ctx = null;
		
		try {
			ctx = new InitialDirContext(env);

			// check if ou=addressbook exists
				SearchControls ctls = new SearchControls();
				ctls.setSearchScope(SearchControls.ONELEVEL_SCOPE);
				NamingEnumeration enm = ctx.search("", "ou=addressbook", ctls);
				boolean abExists = enm.hasMore();
				enm.close();

				if (!abExists) {
					BasicAttributes attrs = new BasicAttributes();
					Attribute classes = new BasicAttribute("objectclass");
					// add object classes
					classes.add("top");
					classes.add("organizationalUnit");
					// add attributes
					attrs.put(classes);
					attrs.put("ou", "addressbook");
					// write subcontext to LDAP
					ctx.createSubcontext(ldapserver + "ou=addressbook,"+ baseDN, attrs);

					// ...and we add ou=public,ou=addressbook as well

					// remove all current object classes
					classes.clear();
					// add object classes for addressbook
					classes.add("top");
					classes.add("organizationalUnit");
					// remove attributes from above
					
					// add new attributes
					attrs.put(classes);
					attrs.put("ou", "public");

					// write subcontext to LDAP
					ctx.createSubcontext(ldapserver + "ou=public,ou=addressbook," + baseDN, attrs);
				}				
				
				ctx.close();
			} catch (NamingException e) {
				e.printStackTrace();
			} finally {
				ctx.close();
			}
		
	}
	
	private static void ldapCleanup (Hashtable env, String ldapserver, String baseDN) throws NamingException {
		DirContext ctx = null;
		try {
			ctx = new InitialDirContext(env);

				SearchControls ctls = new SearchControls();
				ctls.setSearchScope(SearchControls.ONELEVEL_SCOPE);
				
				// delete all old users and their contacts
				NamingEnumeration enm = ctx.search("ou=addressbook", "uid=*", ctls);
				NamingEnumeration enm2;
				while (enm.hasMore()) {
					SearchResult sr = (SearchResult) enm.next();
					Attributes attr = sr.getAttributes();
					// delete all contacts the user has right now
					ctls.setReturningAttributes(new String[] {"uid"});
					enm2 = ctx.search("ou=addressbook,uid="+ attr.get("uid").get() + ",ou=addressbook", "uid=*", ctls);
					// delete all contacts
					while (enm2.hasMore()) {
						SearchResult sr2 = (SearchResult) enm2.next();
						Attributes attr2 = sr2.getAttributes();
						ctx.destroySubcontext("uid=" + attr2.get("uid").get() + ",ou=addressbook,uid="+ attr.get("uid").get() + ",ou=addressbook");
					}
					// delete ou=addressbook
					ctx.destroySubcontext("ou=addressbook,uid="+ attr.get("uid").get() + ",ou=addressbook");
					// delete uid=user
					ctx.destroySubcontext("uid="+ attr.get("uid").get() + ",ou=addressbook");
					enm2.close();
				}
				
				// delete all old public addressbooks and their contacs
				enm = ctx.search("ou=public,ou=addressbook", "uid=*", ctls);
				while (enm.hasMore()) {
					SearchResult sr = (SearchResult) enm.next();
					Attributes attr = sr.getAttributes();
					// delete all contacts the addrssbook has right now
					ctls.setReturningAttributes(new String[] {"uid"});
					enm2 = ctx.search("ou=addressbook,uid="+ attr.get("uid").get() + ",ou=public,ou=addressbook", "uid=*", ctls);
					// delete all contacts
					while (enm2.hasMore()) {
						SearchResult sr2 = (SearchResult) enm2.next();
						Attributes attr2 = sr2.getAttributes();
						ctx.destroySubcontext("uid=" + attr2.get("uid").get() + ",ou=addressbook,uid="+ attr.get("uid").get() + ",ou=public,ou=addressbook");
					}
					// delete ou=addressbook
					ctx.destroySubcontext("ou=addressbook,uid="+ attr.get("uid").get() + ",ou=public,ou=addressbook");
					// delete uid=publicfolder
					ctx.destroySubcontext("uid="+ attr.get("uid").get() + ",ou=public,ou=addressbook");
					enm2.close();
				}
	
				enm.close();

				ctx.close();
			} catch (NamingException e) {
				e.printStackTrace();
			} finally {
				ctx.close();
			}
		
	}
	
	private static void writePrivateContacts(Hashtable env, String ldapserver, String baseDN, MapiConnect mapi) throws NamingException {
		
		DirContext ctx = null;
		// get all private contacts
		Users[] user = mapi.getPrivateContacts();
		
		try {
			ctx = new InitialDirContext(env);

			// check for each user, if uid=USER,ou=addressbook is already there

			for (int i = 0; i < user.length; i++) {
				SearchControls ctls = new SearchControls();
				ctls.setSearchScope(SearchControls.ONELEVEL_SCOPE);
				NamingEnumeration enm = ctx.search("ou=addressbook", "uid=" + user[i].getUsername(), ctls);
				boolean userExists = enm.hasMore();
				enm.close();

				// user does not exist yet. therefore we add the user
				if (!userExists) {
					BasicAttributes attrs = new BasicAttributes();
					Attribute classes = new BasicAttribute("objectclass");
					// add object classes
					classes.add("top");
					classes.add("inetOrgPerson");
					// add attributes
					attrs.put(classes);
					attrs.put("uid", user[i].getUsername());
					attrs.put("sn", user[i].getUsername());
					attrs.put("cn", user[i].getUsername());
					// write subcontext to LDAP
					ctx.createSubcontext(ldapserver + "uid="+ user[i].getUsername() + ",ou=addressbook,"+ baseDN, attrs);

					// add ou=addressbook,uid=USER,ou=addressbook

					// remove all current object classes
					classes.clear();
					// add object classes for addressbook
					classes.add("top");
					classes.add("organizationalUnit");
					// remove attributes from above
					attrs.remove("uid");
					attrs.remove("sn");
					attrs.remove("cn");
					// add new attributes
					attrs.put(classes);
					attrs.put("ou", "addressbook");

					// write subcontext to LDAP
					ctx.createSubcontext(ldapserver + "ou=addressbook,uid="	+ user[i].getUsername() + ",ou=addressbook," + baseDN, attrs);

				}

				// at this point the user and the his addressbook exists in LDAP
				
				// get contacts of each user
				Contact[] contact = user[i].getContacts();
				
				// add every single contact to LDAP
				for (int j = 0; j < contact.length; j++) {
					BasicAttributes attrNewContact = getBasicAttribute(contact[j]);
					try {
						ctx.createSubcontext(ldapserver + "uid=" + contact[j].getUid() + ",ou=addressbook,uid=" + user[i].getUsername() + ",ou=addressbook," + baseDN, attrNewContact);
					} catch (Exception e) {
						System.out.println("Could not add contact " + attrNewContact.get("cn").get().toString() + " from user " + user[i].getUsername());
						System.out.println(e.getLocalizedMessage());
					}

				}
			}
			ctx.close();
		} catch (NamingException e) {
			e.printStackTrace();
		} finally {
			ctx.close();
		}
	}

	private static void writePublicContacts(Hashtable env, String ldapserver, String baseDN, MapiConnect mapi) throws NamingException {

		DirContext ctx = null;
		// get all private contacts
		Folders[] folder = mapi.getPublicContacts();
		if (folder != null) {
			try {
				ctx = new InitialDirContext(env);

				// check for each folder, if uid=FOLDER,ou=public,ou=addressbook is
				// already there

				for (int i = 0; i < folder.length; i++) {

					SearchControls ctls = new SearchControls();
					ctls.setSearchScope(SearchControls.ONELEVEL_SCOPE);
					NamingEnumeration enm = ctx.search("ou=public,ou=addressbook", "uid="	+ folder[i].getFoldername(), ctls);
					boolean folderExists = enm.hasMore();
					enm.close();

					// folder does not exist yet. therefore we add the user
					if (!folderExists) {
						BasicAttributes attrs = new BasicAttributes();
						Attribute classes = new BasicAttribute("objectclass");
						// add object classes
						classes.add("top");
						classes.add("inetOrgPerson");
						// add attributes
						attrs.put(classes);
						attrs.put("uid", folder[i].getFoldername());
						attrs.put("sn", folder[i].getFoldername());
						attrs.put("cn", folder[i].getFoldername());
						// write subcontext to LDAP
						ctx.createSubcontext(ldapserver + "uid="+ folder[i].getFoldername()	+ ",ou=public,ou=addressbook," + baseDN, attrs);

						//add ou=addressbook,uid=FOLDER,ou=public,ou=addressbook

						// remove all current object classes
						classes.clear();
						// add object classes for addressbook
						classes.add("top");
						classes.add("organizationalUnit");
						// remove attributes from above
						attrs.remove("uid");
						attrs.remove("sn");
						attrs.remove("cn");
						// add new attributes
						attrs.put(classes);
						attrs.put("ou", "addressbook");

						// write subcontext to LDAP
						ctx.createSubcontext(ldapserver + "ou=addressbook,uid="	+ folder[i].getFoldername() + ",ou=public,ou=addressbook," + baseDN, attrs);

					}

					// at this point the public folder and the its addressbook exists in LDAP

					// get contacts of each folder
					Contact[] contact = folder[i].getContacts();

					// add every single contact to LDAP
					for (int j = 0; j < contact.length; j++) {
						BasicAttributes attrNewContact = getBasicAttribute(contact[j]);
						try {
							ctx.createSubcontext(ldapserver + "uid=" + contact[j].getUid() + ",ou=addressbook,uid="	+ folder[i].getFoldername()	+ ",ou=public,ou=addressbook," + baseDN, attrNewContact);
						} catch (Exception e) {
							System.out.println("Could not add contact " + attrNewContact.get("cn").get().toString() + " from public addressbook " + folder[i].getFoldername());
							System.out.println(e.getLocalizedMessage());
						}
					}
				}
				ctx.close();
			} catch (NamingException e) {
				e.printStackTrace();
			} finally {
				ctx.close();
			} // end try
		} // end if
	}

	private static BasicAttributes getBasicAttribute(Contact contact) throws NamingException {

		BasicAttributes attrNewContact = new BasicAttributes();
		Attribute classes = new BasicAttribute("objectclass");
		// add object classes
		classes.add("top");
		classes.add("inetOrgPerson");
		classes.add("zarafa-user");
		// add attributes
		attrNewContact.put(classes);

		// add attributes
		attrNewContact.put("uid", contact.getUid());
		attrNewContact.put("cn", (contact.getDisplayName() == null || contact.getDisplayName().equals("")) ? "": contact.getDisplayName());
		attrNewContact.put("sn", (contact.getSurname() == null || contact.getSurname().equals("")) ? " ": contact.getSurname());
		attrNewContact.put("givenName", (contact.getGivenname() == null || contact.getGivenname().equals("")) ? " "	: contact.getGivenname());
		attrNewContact.put("mail", (contact.getEmail1Address() == null || contact.getEmail1Address().equals("")) ? " " : contact.getEmail1Address());

		if (!(contact.getEmail2Address() == null || contact.getEmail2Address().equals("")))
			attrNewContact.put("zarafaAliases", contact.getEmail2Address());

		if (!(contact.getEmail3Address() == null || contact.getEmail3Address().equals("")))
			attrNewContact.put("zarafaAliases", contact.getEmail3Address());

		if (!(contact.getMobile() == null || contact.getMobile().equals("")))
			attrNewContact.put("mobile", contact.getMobile());
		
		if (!(contact.getTelephoneNumber() == null || contact.getTelephoneNumber().equals(""))) 
			attrNewContact.put("telephoneNumber", contact.getTelephoneNumber());
		
		if (!(contact.getFacsimileTelephoneNumber() == null || contact.getFacsimileTelephoneNumber().equals("")))
			attrNewContact.put("facsimileTelephoneNumber", contact.getFacsimileTelephoneNumber());
		
		if (!(contact.getO() == null || contact.getO().equals("")))
			attrNewContact.put("o", contact.getO());

		if (!(contact.getOu() == null || contact.getOu().equals("")))
			attrNewContact.put("ou", contact.getOu());

		if (!(contact.getStreet() == null || contact.getStreet().equals("")))
			attrNewContact.put("street", contact.getStreet());

		if (!(contact.getPostalAddress() == null || contact.getPostalAddress().equals("")))
			attrNewContact.put("postalAddress", contact.getPostalAddress());

		if (!(contact.getPostalCode() == null || contact.getPostalCode().equals("")))
			attrNewContact.put("postalCode", contact.getPostalCode());

		if (!(contact.getHomePostalAddress() == null || contact.getHomePostalAddress().equals("")))
			attrNewContact.put("homePostalAddress", contact.getHomePostalAddress());

		if (!(contact.getTitle() == null || contact.getTitle().equals("")))
			attrNewContact.put("title", contact.getTitle());

		if (!(contact.getHomePhone() == null || contact.getHomePhone().equals(""))) {
			attrNewContact.put("homePhone", contact.getHomePhone());
		}

		if (!(contact.getDescription() == null || contact.getDescription().equals("")))
			attrNewContact.put("description", contact.getDescription());

		if (!(contact.getPager() == null || contact.getPager().equals("")))
			attrNewContact.put("pager", contact.getPager());

		if (!(contact.getEmployeeType() == null || contact.getEmployeeType().equals("")))
			attrNewContact.put("employeeType", contact.getEmployeeType());

		return attrNewContact;

	}

}